import json
import logging
from multiprocessing import Process, Queue

import redis

from inventory_provider import constants
from inventory_provider import snmp
from inventory_provider import juniper


def get_router_interfaces_q(router, params, q):
    threading_logger = logging.getLogger(constants.THREADING_LOGGER_NAME)
    threading_logger.debug("[ENTER>>] get_router_interfaces_q: %r" % router)
    q.put(list(snmp.get_router_interfaces(router, params)))
    threading_logger.debug("[<<EXIT]  get_router_interfaces_q: %r" % router)


def ssh_exec_commands_q(hostname, ssh_params, commands, q):
    threading_logger = logging.getLogger(constants.THREADING_LOGGER_NAME)
    threading_logger.debug("[ENTER>>] exec_router_commands_q: %r" % hostname)
    q.put(list(juniper.ssh_exec_commands(hostname, ssh_params, commands)))
    threading_logger.debug("[<<EXIT] exec_router_commands_q: %r" % hostname)


def get_router_details(router, params):

    threading_logger = logging.getLogger(constants.THREADING_LOGGER_NAME)

    threading_logger.debug("[ENTER>>]get_router_details: %r" % router)

    commands = list(juniper.shell_commands())

    snmpifc_proc_queue = Queue()
    snmpifc_proc = Process(
        target=get_router_interfaces_q,
        args=(router, params, snmpifc_proc_queue))
    snmpifc_proc.start()

    commands_proc_queue = Queue()
    commands_proc = Process(
        target=ssh_exec_commands_q,
        args=(
            router["hostname"],
            params["ssh"],
            [c["command"] for c in commands],
            commands_proc_queue))
    commands_proc.start()

    threading_logger.debug("waiting for commands result: %r" % router)
    command_output = commands_proc_queue.get()
    assert len(command_output) == len(commands)

    r = redis.StrictRedis(
        host=params["redis"]["hostname"],
        port=params["redis"]["port"])
    for c, o in zip(commands, command_output):
        if c["key"]:
            r.hset(
                name=router["hostname"],
                key=c["key"],
                value=json.dumps(c["parser"](o)))
    commands_proc.join()
    threading_logger.debug("... got commands result & joined: %r" % router)

    threading_logger.debug("waiting for snmp ifc results: %r" % router)
    r.hset(
        name=router["hostname"],
        key="snmp-interfaces",
        value=json.dumps(snmpifc_proc_queue.get()))
    snmpifc_proc.join()
    threading_logger.debug("... got snmp ifc result & joined: %r" % router)

    threading_logger.debug("[<<EXIT]get_router_details: %r" % router)


def update_network_details(params):

    threading_logger = logging.getLogger(constants.THREADING_LOGGER_NAME)

    processes = []
    for r in params["routers"]:
        p = Process(target=get_router_details, args=(r, params))
        p.start()
        processes.append({"router": r, "process": p})

    result = {}
    for p in processes:
        threading_logger.debug(
            "waiting for get_router_details result: %r" % p["router"])
        p["process"].join()
        threading_logger.debug(
            "got result and joined get_router_details proc: %r" % p["router"])

    return result


def load_network_details(redis_params):

    r = redis.StrictRedis(
        host=redis_params["hostname"],
        port=redis_params["port"])

    result = {}
    for hostname in r.keys():
        host = {}
        for key in r.hkeys(name=hostname):
            host[key.decode("utf-8")] = json.loads(
                r.hget(hostname, key).decode("utf-8")
            )
        result[hostname.decode("utf-8")] = host

    return result
